'''
Adjustment Method 调整方法
'''
'''
OLS regrssion
'''

from numpy import ndarray

from statsmodels.api import OLS, add_constant
import numpy as np
from numpy.linalg import inv
from scipy.stats import t

def ols(y: ndarray, x:ndarray, constant: bool=True) :
    '''
    The function is for OLS regression, which is equal to the OLS module in package *statsmodels*. 
    input :
        y (ndarray): The dependent variable.
        x (ndarray): The explanatory variable.
        constant (boolean): add constant or not in OLS model. The default is *True*.

    output :
        result (OLSRegressResult): The result of the regression.
    '''
    
    if constant == True :
        result = OLS(y,add_constant(x)).fit()
    elif constant == False :
        result = OLS(y,x).fit()
    
    return result

'''
White Estimate of Variance 
'''
def white(y: ndarray, X:ndarray, **kwargs) ->ndarray:
    '''
    White Estimate of Variance
    X(r,c): r is smaple number, c is variable number
    S0 is coviarance matrix of residue
    The Variance V is estiamted as V_ols=T(X'X)^(-1)S0(X'X)^(-1)
    The white estimation of S0 is S0=1/T*X'(X*resid_sqr)
    x_i=[xt1,xt2,...,xtk+1]'

    input :
        y (ndarray): The dependent variable.
        X (ndarray): The explanatory variable.

    output :
        V_ols (ndarray): The white variance estimate

    Example:
    # continue the previous code
    import numpy as np
    from statsmodels.api import add_constant

    X = np.random.normal(loc=0.0, scale=1.0, size=(2000,10))
    b = np.random.uniform(low=0.1, high=1.1, size=(10,1))
    e = np.random.normal(loc=0.0, scale=1.0, size=(2000,1))
    y = X.dot(b) + e + 1.0

    re = white(y, X, constant=True)
    np.set_printoptions(formatter={'float':'{:0.3f}'.format})
    print(re*100)
    X = add_constant(X)
    r, c = np.shape(X)
    print('\n', 1/r*X.T.dot(X).dot(re).dot(X.T.dot(X)))
    '''

    r, c = np.shape(X)
    result = ols(y, X, **kwargs)
    resid = result.resid
    resid_sqr = np.diag(resid**2)
    if kwargs['constant'] == True:
        X = add_constant(X)
    S0 = 1 / r * X.T.dot(resid_sqr.dot(X))
    temp = X.T.dot(X)
    V_ols = r*inv(temp).dot(S0).dot(inv(temp))
    
    return V_ols

'''
Newey_West Estimate of Variance
'''
def newey_west(y:ndarray, X:ndarray, J: int=None, **kwargs) -> ndarray:
    '''
    Newey West Estimate of Variance
    X(r,c): r is smaple number, c is variable number
    S0 is coviarance matrix of residue
    The Variance V is estiamted as V_ols=T(X'X)^(-1)S0(X'X)^(-1)
    The white estimation of S0 is S0=1/T*X'(X*resid_sqr)+1/T*sum_j(sum_t(w_j*e_t*e_t-j(x_tx'_t-j+xt_-jx'_t)))
    x_i=[xt1,xt2,...,xtk+1]'

    input :
        y (ndarray): The dependent variable.
        X (ndarray): The explanatory variable.
        J (int): The lag.

    output :
        V_ols (ndarray): The Newey-West variance estimate.
    
    Example:
    # continue the previous code
    import numpy as np
    from statsmodels.stats.sandwich_covariance import cov_hac

    X = np.random.normal(loc=0.0, scale=1.0, size=(10000,10))
    b = np.random.uniform(low=0.1, high=1.1, size=(10,1))
    e = np.random.normal(loc=0.0, scale=1.0, size=(10000,1))
    y = X.dot(b) + e + 1.0

    re = newey_west(y, X, constant=False)
    np.set_printoptions(formatter={'float':'{:0.2f}'.format})
    print(re)
    # X = add_constant(X)
    r, c = np.shape(X) 
    print('\n', 1/r*X.T.dot(X).dot(re).dot(X.T.dot(X)))
    result = ols(y, X, constant=False)
    print('\n', cov_hac(result))
    print('\n', 1/r*X.T.dot(X).dot(cov_hac(result)).dot(X.T.dot(X)))
    '''

    r, c = np.shape(X)
    result = ols(y, X, **kwargs)
    resid = result.resid
    
    if J != None :
        J = J
    elif J == None :
        J = int(np.floor(4*(r/100)**(2/9)))
    
    if kwargs['constant'] == True:
        X = add_constant(X)
    r_X, c_X = np.shape(X)
    temp_cor = np.zeros((c_X, c_X))
    X = np.matrix(X)
    for j in range(1, J+1) :
        # the weight w_j= 1- j/(+J), j start from 1
        # in python however the index starts from 0, and thus j+1 is used 
        w_j = 1 - j / (1 + J)
        for t in range(j+1, r+1) :
            temp_cor = temp_cor + (1 / r) * w_j * resid[t-1] * resid[t-j-1] * \
                       (X[t-1, :].T.dot(X[t-j-1, :]) + X[t-j-1,:].T.dot(X[t-1, :]))

    resid_sqr = np.diag(resid**2)

    temp_var =  1 / r * X.T.dot(resid_sqr.dot(X))    
    S0 = temp_var + temp_cor
    temp = X.T.dot(X)
    V_ols = r * inv(temp).dot(S0).dot(inv(temp))

    return V_ols

'''
T test after adjustment of White
'''
def white_t(y:ndarray, X:ndarray, params: ndarray=None, side: str='Two', **kwargs) :
    '''
    White t test based on White Variance
    This function constructs t-test based on White variance estimate.

    input :
        y (ndarray): The dependent variable.
        X (ndarray): The explanatory variable.
        params (ndarray): Already have parameters or not. The default is *None*.

    output :
        t_value (ndarray): The t-value of parameters.
        p_value (ndarray): The p-value of parameters.
    
    Example:
    import numpy as np

    X = np.random.normal(loc=0.0, scale=1.0, size=(10000,10)) 
    b = np.random.uniform(low=-0.5, high=0.5, size=(10,1))
    e = np.random.normal(loc=0.0, scale=1.0, size=(10000,1))
    y = X.dot(b) + e + 1.0

    re = white_t(y, X, constant=False)
    np.set_printoptions(formatter={'float':'{:0.2f}'.format})
    print('t_value : ', re[0], '\np_value : ', re[1])
    print('b :', b.T)
    '''

    diagonal = np.diagonal(white(y, X, **kwargs))
    standard_error = diagonal**0.5
    if params == None :
        result = ols(y, X, **kwargs)
        params = result.params
    t_value = params / standard_error
    r, c = np.shape(X)
    if kwargs['constant'] == True :
        freedom = r - c - 1
    else :
        freedom = r - c
    
    if side == 'One':
        p_value = 1- t.cdf(np.abs(t_value), freedom)
    elif side == 'Two':
        p_value = 2 * (1- t.cdf(np.abs(t_value), freedom))
    
    return t_value, p_value

'''
T test after adjustment of Newey West
'''
def newey_west_t(y:ndarray, X:ndarray, params: ndarray=None, side: str='Two', **kwargs) :
    '''
    Newey_West t test based on Newey West Variancce 
    This function constructs t-test based on Newey-West variance estimate.
    input :
        y (ndarray): The dependent variable.
        X (ndarray): The explanatory variable.
        params (ndarray): Already have parameters or not. The default is None.

    output :
        t_value (ndarray): The t-value of parameters.
        p_value (ndarray): The p-value of parameters.
    
    Example:
    import numpy as np

    X = np.random.normal(loc=0.0, scale=1.0, size=(10000,10))
    b = np.random.uniform(low=-0.5, high=0.5, size=(10,1))
    e = np.random.normal(loc=0.0, scale=1.0, size=(10000,1))
    y = X.dot(b) + e + 1.0

    re = newey_west_t(y, X, constant=True)
    np.set_printoptions(formatter={'float':'{:0.2f}'.format})
    print('t_value : ', re[0], '\np_value : ', re[1])
    print('b :', b.T)
    '''

    diagnal = np.diagonal(newey_west(y, X, **kwargs))
    standard_error = diagnal**0.5
    if params is not None:
        t_value = params / standard_error
    else:
        constant = kwargs['constant']
        result = ols(y, X, constant=constant)
        params = result.params
        t_value = params / standard_error
    
    r, c = np.shape(X)
    if kwargs['constant'] == True :
        freedom = r - c - 1
    else :
        freedom = r - c
    
    if side == 'One':
        p_value = 1- t.cdf(np.abs(t_value), freedom)
    elif side == 'Two':
        p_value = 2 * (1- t.cdf(np.abs(t_value), freedom))

    return t_value, p_value

        


'''
Fama-Macbeth Regerssion Fama-Macbeth 回归
Following Two step 分为两步
1. specify the model and take cross-sectional regression 确定模型，进行截面回归 
2. take the time-series average of regress coefficient 对系数在时间序列上取平均

For more academic reference: 
Empirical Asset Pricing: The Cross Section of Stock Returns. Bali, Engle, Murray. 2016.
'''

from numpy import ndarray

import numpy as np
import statsmodels.api as sm
import scipy.stats as sts
from prettytable import PrettyTable
import pandas as pd
import matplotlib.pyplot as plt

from .portfolio_analysis import Bivariate
from .adjust import newey_west_t


class Fama_macbeth_regress():
    '''
    fama_macbeth_regress follwing two steps
    Need Package: numpy,statsmodels,scipy,prettytable
    '''
    def __init__(self, sample:ndarray) -> None:
        '''
        input: 
            sample (ndarray/DataFrame): data used for analysis 用于回归数据
            The strucure of the sample：
                the first column is dependent variable/ Test Portfolio Return
                the second to the last-1 columns is independent variable/ factor loadings
                the last coolumn is time label
        '''

        if type(sample).__name__ == 'DataFrame':
            self.sample = np.array(sample)
            self._input_type = 'DataFrame'
            self._columns = list(sample.columns)
        elif type(sample).__name__ == 'ndarray':
            self.sample = sample
            self._input_type = 'ndarray'
        # self.time_series_average()
    
    def divide_by_time(self, sample:ndarray):
        '''
        This function group the sample by time.
        input :
            sample (ndarray/DataFrame): The data for analysis in the __init__ function.

        output :
            groups_by_time (list): The sample grouped by time.
        '''
        
        # extracting the sample time label 提取样板时间戳
        time = np.sort(np.unique(sample[:,-1]))
        # group the sample by time 按时间对样本中的数据进行分类
        groups_by_time = list()
        for i in range(len(time)):
            groups_by_time.append(sample[np.where(sample[:,-1]==time[i])])
        
        # return grouped sample 返回被分类的样本组
        return groups_by_time
            
    def cross_sectional_regress(self, add_constant: bool=True, normalization: bool=True, **kwargs):
        '''
        The #1 step of Fama-Macbeth Regression Fama-Macbeth 回归第一步
        take the cross_sectional regress for each period 在每个时间段分别进行截面数据回归
        input : **kwargs: 
                add_constant (boolean): whether add_constant when take CS regression 截面回归时是否添加常数
        output: 
            parameters (ndarray): The regression coefficient/factor risk premium, whose rows are the group coefficient and columns are regression variable.
            tvalue (ndarray): t value for the coefficient.
            rsquare (list): The r-square.
            adjrsq (list):  The adjust r-square.
            n (list): The sample quantity in each group.
        '''
        
        # get sample shape 获取样本行列数
        r,c = np.shape(self.sample)
        # get sample groups split by time 获取分组样本 
        groups = self.divide_by_time(self.sample)
        # initiate the variable 初始化变量
        
        if add_constant == True :
            params = np.zeros((len(groups), c-1))
            tvalue = np.zeros((len(groups), c-1))
        elif add_constant == False :
            params = np.zeros((len(groups), c-2))
            tvalue = np.zeros((len(groups), c-2))
        rsquare = list()
        adjrsq = list()
        n = list()
        
        # cross section regression by groups 分组截面回归
        for i in range(len(groups)) :
            # group size 组的大小
            r, c = np.shape(groups[i])
            temp_group = groups[i]
            # regression 回归
            if add_constant == True :
                if normalization == True:
                    temp_endog = (temp_group[:, 0]-np.mean(temp_group[:, 0]))/np.std(temp_group[:, 0])
                    temp_exog = (temp_group[:, 1:-1]-np.mean(temp_group[:, 1:-1], axis=0))/(np.var(temp_group[:, 1:-1], axis=0)**0.5)
                    result = sm.OLS(temp_endog.astype(float), sm.add_constant(temp_exog.astype(float))).fit()
                
                elif normalization == False:
                    result = sm.OLS(temp_group[:, 0].astype(float), sm.add_constant(temp_group[:, 1:-1].astype(float))).fit()    
                
                self.constant = True
            elif add_constant == False :
                if normalization == True:
                    temp_endog = (temp_group[:, 0]-np.mean(temp_group[:, 0]))/np.std(temp_group[:, 0])
                    temp_exog = (temp_group[:, 1:-1]-np.mean(temp_group[:, 1:-1], axis=0))/(np.var(temp_group[:, 1:-1], axis=0)**0.5)
                    result = sm.OLS(temp_endog.astype(float), temp_exog.astype(float)).fit()
  
                elif normalization == False:
                    result = sm.OLS(temp_group[:, 0].astype(float), temp_group[:, 1:-1].astype(float)).fit()
                
                self.constant = False
            # params of each group
            params[i,:] = result.params
            # t_vlaue of each group
            tvalue[i,:] = result.tvalues
            # R square of each group
            rsquare.append(result.rsquared)
            # adjust R square of each group
            adjrsq.append(result.rsquared_adj)
            # sample number of each group
            n.append(r)
        
        return params, tvalue, rsquare, adjrsq, n
    
    def time_series_average(self, maxlag: int=12, **kwargs) :
        '''
        The #2 step of Fama-French regression Fama-French 回归第二步 
        Time series average for cross section regression  对截面数据在时间上取均值
        '''
        
        # the params of cross-section regression 截面回归相关参数 
        self.result_cross = self.cross_sectional_regress(**kwargs)
        # the average of the regression coefficeint 参数均值
        self.coefficient_average = np.mean(self.result_cross[0], axis=0)
        # t_test for the regression coefficient 回归系数T检验
        if maxlag == 0.0:
            self.tvalue = sts.ttest_1samp(self.result_cross[0], 0.0, axis=0)[0]
        else:
            r, c = np.shape(self.result_cross[0])
            self.tvalue = np.zeros((c))

            for i in range(c):
                temp_result = self.result_cross[0][:, i]
                temp_result = temp_result[~np.isnan(temp_result)]
                temp_one = np.ones((len(temp_result), 1))
                tvalue, pvalue = newey_west_t(y=temp_result, X=temp_one, J=maxlag, constant=False)
                self.tvalue[i] = tvalue

        # the average of the R square R方均值
        self.rsquare_average = np.mean(self.result_cross[2])
        # the average of the adjust R square 调整R方均值
        self.adjrsq_average = np.mean(self.result_cross[3])
        # the average sample numbers in each group 各组样本均值
        self.n_average = np.mean(self.result_cross[4])
        # 设置时间
        self._time = np.sort(np.unique(self.sample[:,-1]))
        # print the result 打印结果
        print('para_average:', self.coefficient_average)
        print('tvalue:', self.tvalue)
        print('R:', self.rsquare_average)
        print('ADJ_R:', self.adjrsq_average)
        print('sample number N:', self.n_average)
        # print(np.vstack([self.para_average,self.tvalue,np.array(self.rsquare_average)]))
    
    def fit(self, **kwargs) :
        '''
        Fit model 估计模型
        run function: time_series_average 运行函数： time_series_average()
        
        Example:
        import numpy as np
        from fama_macbeth import Fama_macbeth_regress
    
        # construct sample
        year=np.ones((3000,1),dtype=int)*2020
        for i in range(19):
            year=np.append(year,(2019-i)*np.ones((3000,1),dtype=int))
        character=np.random.normal(0,1,(2,20*3000))
        # print('Character:',character)
        ret=np.array([-0.5,-1]).dot(character)+np.random.normal(0,1,20*3000)
        sample=np.array([ret,character[0],character[1],year]).T    
        # print('Sample:',sample)
        # print(sample.shape)

        model = Fama_macbeth_regress(sample) 
        result = model.fit(add_constant=False)
        print(result)
        '''

        self.time_series_average(**kwargs) 

    def summary_by_time(self):
        '''
        summary the cross-section result of each time 总结每一时刻的结果
        package needed : prettytable
        '''
        
        r,c = np.shape(self.sample)
        time = np.sort(np.unique(self.sample[:, -1]))
        result_cross = self.result_cross
        
        table = PrettyTable()
        table.add_column('Year', time)
        table.add_column('Param', result_cross[0])
        table.add_column('R Square', np.round(result_cross[2], decimals=2))
        table.add_column('Adj R Square', np.round(result_cross[3], decimals=2))
        table.add_column('Sample Number', result_cross[4])
        
        # coefficient decimals: 3, R & adj_R decimals: 2
        np.set_printoptions(formatter={'float':'{:0.3f}'.format})
        print(table)
        
    def summary(self,charactername:list =None) :
        '''
        summary the time-series average 总结时间序列平均
        input :
            charactername : The factors' name in the cross-section regression model.
        '''
        
        r,c = np.shape(self.sample)
        table = PrettyTable()
        if charactername == None :
            charactername = list()
            for i in range(c-2):
                charactername.extend(str(i+1))
        if self.constant == True and self._input_type == 'ndarray' :
            table.field_names = ['Intercept','Intercept Tvalue'] + ['Param','Param Tvalue'] + ['Average R','Average adj R','Average n']
            table.add_row([np.around(self.coefficient_average[0], decimals=4), np.around(self.tvalue[0], decimals=3), self.coefficient_average[1:],\
                           self.tvalue[1:], np.around(self.rsquare_average, decimals=3), np.around(self.adjrsq_average, decimals=3),\
                           self.n_average])
        elif self.constant == True and self._input_type == 'DataFrame' :
            table.field_names = ['Intercept'] + self._columns[1:-1] + ['Average R','Average adj R','Average n']
            table.add_row(np.around([self.coefficient_average[0]] + list(self.coefficient_average[1:]) + [np.around(self.rsquare_average, decimals=3), np.around(self.adjrsq_average, decimals=3),\
                           np.around(self.n_average, decimals=2)], decimals=4))
            table.add_row([np.around(self.tvalue[0], decimals=3)] + list(np.around(self.tvalue[1:], decimals=3)) + ['-', '-', '-'])

        elif self.constant == False and self._input_type == 'ndarray' :
            table.field_names = ['Param','Param Tvalue']+['Average R','Average adj R','Average n']
            table.add_row([self.coefficient_average, self.tvalue, np.around(self.rsquare_average, decimals=3), np.around(self.adjrsq_average, decimals=3),\
                           self.n_average])
        elif self.constant == False and self._input_type == 'DataFrame' :
            table.field_names = ['Intercept'] + self._columns[1:-1] + ['Average R','Average adj R','Average n']
            table.add_row(['-'] + list(np.around(self.coefficient_average, decimals=4)) + list([np.around(self.rsquare_average, decimals=3), np.around(self.adjrsq_average, decimals=3),\
                           np.around(self.n_average, decimals=2)]))
            table.add_row(['-'] + list(np.around(self.tvalue, decimals=3)) + ['-', '-', '-'])

        
        print('\n')
        print(table)

    def plot(self, figsize: tuple =(14,7), together:bool =False, window: int = None, select:list =None, **kwargs):
        '''
        plot the risk premium
        '''
        if self._columns[1:-1]:
            fac = self._columns[1:-1]
        else: 
            if self.constant == False:
                fac = ['factor' + str(i) for i in range(len(self.coefficient_average))]
            elif self.constant == True:
                fac = ['factor' + str(i) for i in range(len(self.coefficient_average)-1)]
            else: raise IOError

        if self.constant == False:
            temp_data = pd.DataFrame(self.result_cross[0], columns=fac, index=self._time)
        elif self.constant == True:
            temp_data = pd.DataFrame(self.result_cross[0][:, 1:], columns=fac, index=self._time)
        
        select_list = list()
        if select is not None:
            for i in range(len(select)):
                if type(select[i]).__name__ == 'str':
                    select_list.append(select[i])
                else:
                    select_list.append('factor' + str(select[i]))
            
            temp_data = temp_data[select_list]
            fac = temp_data.columns

        fig, ax = plt.subplots(len(fac), 1, figsize=figsize, dpi=500)
        for i in range(len(fac)):
            temp_data.iloc[:, i].plot(ax=ax[i], **kwargs)
            ax[i].set_ylabel(fac[i])

        if together == True:
            fig_t, ax_t = plt.subplots(1, 1, figsize=figsize, dpi=500)
            temp_data.plot(ax=ax_t, label=True, **kwargs)

        if window is not None:
            fig_w, ax_w = plt.subplots(len(fac), 1,figsize=figsize, dpi=500)
            for i in range(len(fac)):
                temp_data_w = temp_data.iloc[:, i].rolling(window=window).mean()
                temp_data_w.plot(ax=ax_w[i], ylabel=fac[i], **kwargs)
            
            if together == True:
                fig_wt, ax_wt = plt.subplots(1, 1, figsize=figsize, dpi=500)
                temp_data_w = temp_data.rolling(window=window).mean()
                temp_data_w.plot(ax=ax_wt, label=True, **kwargs)

        plt.show()

        
class Factor_mimicking_portfolio():
    '''
    This module is designed for generating factor mimicking portfolio following the Fama-French (1993) conventions, and then calculating factor risk premium.    
    
    1. Group stocks by two dimensions. One dimension is size, divided by 50% small stocks and 50% big stocks. The other is the factor, divided by 30% tail factor stocks, 30%~70% factor stocks, and 70%~100% factor stocks.
    2. Calculate market value weighted portfolio return and factor risk premium.
       SMB=1/3(S/L+S/M+S/H)-1/3(B/L+B/M+B/H)
       Factor=1/2(S/H+B/H)-1/2(S/L+B/L)
    3. In Fama-French (1993), the factor is book-to-market ratio, and other literatures follow the same way to construct factor mimicking portfolio. The return of each portfolio is represented by the market value weighted portfolio return.  
    '''
    def __init__(self, sample:ndarray, perc_row: list=[0, 50, 100], perc_col: list=[0, 30, 70, 100], percn_row:ndarray =None, percn_col:ndarray =None, weight:bool =True):
        '''
        This function initializes the object.

        input :
            sample (ndarray or DataFrame): data for analysis. The structure of the sample : 
                The first column is dependent variable/ test portfolio return.
                The second is the first factor.
                The third is the second factor.
                The fourth is the timestamp.
                The fifth is the weight.
            perc_row (list or array): The percentiles that divide stocks by the first factor. The **Default** percentile is [0, 50, 100].
            perc_col (list or array): The percentiles that divide stocks by the second factor. The **Default** percentile is [0, 30, 70, 100].
            percn_row (list/array): The percentile that divide stocks by the first factor.
            percn_col (list/array): The percentile that divide stocks by the second factor. 
            weight (array or Series): Whether the portfolio return calculated by weight. The **Default** is True.
        '''
        
        self.sample = sample
        self.perc_row = perc_row
        self.perc_col = perc_col
        self.percn_row = percn_row
        self.percn_col = percn_col
        self.weight = weight
        if type(sample).__name__ == 'DataFrame':
            self._time = np.sort(np.unique(sample.iloc[:, 3])) 
        elif type(sample).__name__ == 'ndarray':
            self._time = np.sort(np.unique(sample[:, 3]))

    def portfolio_return_time(self):
        '''
        Construct portfolio and calculate the average return and difference matrix 
        This function is to construct portfolio and calculate the average return and difference matrix.
        output : 
            diff (ndarray): The differenced portfolio return matrix.
        
        Example:
        TEST Factor_mimicking_portfolio
        construct sample:
        1. 20 Periods
        2. 3000 Observations for each Period
        3. Character negative with return following the return=character*-0.5+sigma where sigma~N(0,1)

        import numpy as np
        from fama_macbeth import Factor_mimicking_portfolio
    
        # construct sample
        year=np.ones((3000,1),dtype=int)*2020
        for i in range(19):
            year=np.append(year,(2019-i)*np.ones((3000,1),dtype=int))
        character=np.random.normal(0, 1, (2, 20*3000))
        weight = np.random.uniform(0, 1, (1, 20*3000))
        ret=np.array([-0.5,-1]).dot(character)+np.random.normal(0,1,20*3000)
        sample=np.array([ret, character[0], character[1], year, weight[0]]).T

        model = Factor_mimicking_portfolio(sample)
        portfolio_return_time = model.portfolio_return_time()
        print('portfolio_return_time:', portfolio_return_time)
        print('portfolio_return_time:', np.shape(portfolio_return_time))
        '''
        
        bi = Bivariate(self.sample)
        if all([self.percn_row is None, self.percn_col is None]):
            bi.fit(perc_row=self.perc_row, perc_col=self.perc_col, weight=self.weight)
        else:
            bi.fit(percn_row=self.percn_row, percn_col=self.percn_col, weight=self.weight)
        diff = bi.difference(bi.average_by_time())
        
        return diff
    
    def portfolio_return(self):
        '''
        Construct factor risk premium
        This function is to construct factor risk premium.
        output :
            return_row : The first factor risk premium. The **Default** is size factor.
            return_col : The second factor risk premium. 
        
        '''
        
        diff = self.portfolio_return_time()
        r, c, n = np.shape(diff)
        
        time = self._time

        return_row = pd.Series(np.mean(diff[-1, :c, :], axis=0), index=pd.to_datetime(time))
        return_col = pd.Series(np.mean(diff[:r, -1, :], axis=0), index=pd.to_datetime(time))
        
        return return_row, return_col
    
    def portfolio_return_horizon(self, period: int, ret: bool=False):
        '''
        Construct horizon pricing factor
        This function is to construct horizon pricing factor. For details, read Horizon Pricing, JFQA, 2016, 51(6): 1769-1793.
        input :
            period (int): The lagged period for constructing factor risk premium return.
            log(boolean): whether use log return.

        output :
        return_row_multi : The first factor risk premium. The DEFAULT is size factor.
        return_col_multi : The second factor premium.

        Example:
        # Continue the previous code
        portfolio_return = model.portfolio_return()
        print('portfolio_return_row: \n', portfolio_return[0])
        print('portfolio_return_row:', np.shape(portfolio_return[0]))
        print('portfolio_return_col: \n', portfolio_return[1])
        print('portfolio_return_col:', np.shape(portfolio_return[1]))
        '''

        diff = self.portfolio_return_time()
        r, c, n =np.shape(diff)

        time = self._time

        return_col_up = pd.Series(np.mean(diff[:-1, 0, :], axis=0), index=pd.to_datetime(time))
        return_col_down = pd.Series(np.mean(diff[:-1, -2, :], axis=0), index=pd.to_datetime(time))
        return_row_up = pd.Series(np.mean(diff[0, :c, :], axis=0), index=pd.to_datetime(time))
        return_row_down = pd.Series(np.mean(diff[-2, :c, :], axis=0), index=pd.to_datetime(time))

        return_row_up_plus = np.log(return_row_up + 1)
        return_row_down_plus = np.log(return_row_down + 1)
        return_col_up_plus = np.log(return_col_up + 1)
        return_col_down_plus = np.log(return_col_down + 1)
        
        if period > 1:
            return_row_up_multi = return_row_up_plus.rolling(window=period).sum()
            return_row_down_multi = return_row_down_plus.rolling(window=period).sum()
            return_col_up_multi = return_col_up_plus.rolling(window=period).sum()
            return_col_down_multi = return_col_down_plus.rolling(window=period).sum()
            
            if ret == False:
                return_row_multi = np.exp(return_row_down_multi) - np.exp(return_row_up_multi)
                return_col_multi = np.exp(return_col_down_multi) - np.exp(return_col_up_multi)
            elif ret == True:
                return_row_multi = return_row_down_multi - return_row_up_multi
                return_col_multi = return_col_down_multi - return_col_up_multi

        elif period == 1:
            if ret == False:
                return_row_multi = np.exp(return_row_down_plus) - np.exp(return_row_up_plus)
                return_col_multi = np.exp(return_col_down_plus) - np.exp(return_col_up_plus)
            elif ret == True:
                return_row_multi = return_row_down_plus - return_row_up_plus
                return_col_multi = return_col_down_plus - return_col_up_plus

        return return_row_multi, return_col_multi




# plot function
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import seaborn as sns
import pandas as pd
import numpy as np

from .portfolio_analysis import ptf_analysis as ptfa

def plot(portfolio:ptfa, pattern:str ='scatter', alpha:bool =False, percentage:bool =True, figsize:tuple =(14,7), pos: float=0.0, **kwargs):
    if type(portfolio).__name__ == 'Univariate':
        plot_type = 'uni'
    elif type(portfolio).__name__ == 'Bivariate':
        plot_type = 'bi'
    
    if alpha == False:
        average_result, ttest = portfolio.return_average_ttest(alpha=alpha)
        alpha_mat = np.zeros(np.shape(average_result))
        alpha_tvalue = np.zeros(np.shape(average_result))
    elif alpha == True: 
        average_result, ttest, alpha_mat, alpha_tvalue = portfolio.return_average_ttest(alpha=alpha)
    
    if percentage == True:
        average_result = 100 * average_result
        alpha_mat = 100 * alpha_mat
    
    if plot_type == 'uni':
        row = len(average_result)
        if len(average_result) >= 2:
            col_names = ['Low'] + [str(i+1) for i in range(len(average_result)-3)] + ['High']
        
        if alpha == True:
            alpha_mat = alpha_mat[:, 0]
            alpha_tvalue = alpha_tvalue[:, 0]
        
        data = {'portfolio': col_names, 'ave_re': average_result[:-1], 'ttest': ttest[0][:-1], 'alpha':alpha_mat[:-1], 'alpha_tvalue':alpha_tvalue[:-1]}
        data = pd.DataFrame(data)

        if pattern == 'scatter':
            if alpha == False:
                fig, ax = plt.subplots(2, 1, figsize=figsize, dpi=500)
                sns.scatterplot(data=data, x='portfolio', y='ave_re', ax=ax[0], **kwargs)
                sns.scatterplot(data=data, x='portfolio', y='ttest', ax=ax[1], **kwargs)
            elif alpha == True:
                fig, ax = plt.subplots(4, 1, figsize=figsize, dpi=500)
                sns.scatterplot(data=data, x='portfolio', y='ave_re', ax=ax[0], **kwargs)
                sns.scatterplot(data=data, x='portfolio', y='ttest', ax=ax[1], **kwargs)
                sns.scatterplot(data=data, x='portfolio', y='alpha', ax=ax[2], **kwargs)
                sns.scatterplot(data=data, x='portfolio', y='alpha_tvalue', ax=ax[3], **kwargs)
        elif pattern == 'bar':
            if alpha == False:
                fig, ax = plt.subplots(2, 1, figsize=figsize, dpi=500)
                sns.barplot(data=data, x='portfolio', y='ave_re', ax=ax[0], **kwargs)
                sns.barplot(data=data, x='portfolio', y='ttest', ax=ax[1], **kwargs)
            elif alpha == True:
                fig, ax = plt.subplots(4, 1, figsize=figsize, dpi=500)
                sns.barplot(data=data, x='portfolio', y='ave_re', ax=ax[0], **kwargs)
                sns.barplot(data=data, x='portfolio', y='ttest', ax=ax[1], **kwargs)
                sns.barplot(data=data, x='portfolio', y='alpha', ax=ax[2], **kwargs)
                sns.barplot(data=data, x='portfolio', y='alpha_tvalue', ax=ax[3], **kwargs)
        elif pattern == 'line':
            if alpha == False:
                fig, ax = plt.subplots(2, 1, figsize=figsize, dpi=500)
                sns.lineplot(data=data, x='portfolio', y='ave_re', ax=ax[0], **kwargs)
                sns.lineplot(data=data, x='portfolio', y='ttest', ax=ax[1], **kwargs)
            elif alpha == True:
                fig, ax = plt.subplots(4, 1, figsize=figsize, dpi=500)
                sns.lineplot(data=data, x='portfolio', y='ave_re', ax=ax[0], **kwargs)
                sns.lineplot(data=data, x='portfolio', y='ttest', ax=ax[1], **kwargs)
                sns.lineplot(data=data, x='portfolio', y='alpha', ax=ax[2], **kwargs)
                sns.lineplot(data=data, x='portfolio', y='alpha_tvalue', ax=ax[3], **kwargs)
        else: raise IOError
    
    elif plot_type == 'bi':
        row, col = np.shape(average_result)
        data = pd.DataFrame(columns=['row', 'col','portfolio', 'ave_re', 'ttest', 'alpha', 'alpha_tvalue'])
        for i in range(row-1):
            for j in range(col-1):
                temp_data = pd.DataFrame({'row':i, 'col':j, 'portfolio': 'Row'+str(i+1)+'Col'+str(j+1), 'ave_re': average_result[i, j], 'ttest': ttest[0][i, j], 'alpha': alpha_mat[i, j], 'alpha_tvalue': alpha_tvalue[i, j]}, index=['Row'+str(i+1)+'Col'+str(j+1)])
                data = pd.concat([data, temp_data])

        if pattern == '3d':
            if alpha == False:
                fig = plt.figure('ave_re', figsize=figsize, dpi=500)
                ax1 = fig.add_subplot(121, projection='3d')
                ax2 = fig.add_subplot(122, projection='3d')
                
                # average_result
                for i in range(len(data['row'].index)):
                    color = (data['ave_re'][i] - min(data['ave_re']))/(max(data['ave_re'])-min(data['ave_re']))
                    ax1.bar3d(data['row'][i], data['col'][i], pos, 1, 1, data['ave_re'][i], color=np.array([1, 0.9, 0.9*color]), **kwargs)
                #ax.bar3d(data['row'], data['col'], 0, 1, 1, data['ave_re'], shade=True, cmap='rainbow')
                ax1.set_xlabel('Row')
                ax1.set_ylabel('Col')
                ax1.set_zlabel('ave_re')
                
                ax1.set_xticks([i for i in range(row)])
                ax1.set_yticks([i for i in range(col)])
                
                if min(data['ave_re']) < 0:
                    zmin = 1.2 * min(data['ave_re'])
                else:
                    zmin = 0.8 * min(data['ave_re'])
                
                if max(data['ave_re']) < 0:
                    zmax = 0.8 * max(data['ave_re'])
                else:
                    zmax = 1.2 * max(data['ave_re'])

                ax1.set_zlim(zmin, zmax)

                # ttest
                for i in range(len(data['row'].index)):
                    color = (data['ttest'][i] - min(data['ttest']))/(max(data['ttest'])-min(data['ttest']))
                    ax2.bar3d(data['row'][i], data['col'][i], pos, 1, 1, data['ttest'][i], color=np.array([1, 0.9, 0.9*color]), **kwargs)
                #ax.bar3d(data['row'], data['col'], 0, 1, 1, data['ave_re'], shade=True, cmap='rainbow')
                ax2.set_xlabel('Row')
                ax2.set_ylabel('Col')
                ax2.set_zlabel('ttest')

                ax2.set_xticks([i for i in range(row)])
                ax2.set_yticks([i for i in range(col)])

                if min(data['ttest']) < 0:
                    zmin = 1.2 * min(data['ttest'])
                else:
                    zmin = 0.8 * min(data['ttest'])
                
                if max(data['ttest']) < 0:
                    zmax = 0.8 * max(data['ttest'])
                else:
                    zmax = 1.2 * max(data['ttest'])

                ax2.set_zlim(zmin, zmax)

            elif alpha == True:
                fig = plt.figure('ave_re', figsize=figsize, dpi=500)
                ax1 = fig.add_subplot(141, projection='3d')
                ax2 = fig.add_subplot(142, projection='3d')
                ax3 = fig.add_subplot(143, projection='3d')
                ax4 = fig.add_subplot(144, projection='3d')

                # average result
                for i in range(len(data['row'].index)):
                    color = (data['ave_re'][i] - min(data['ave_re']))/(max(data['ave_re'])-min(data['ave_re']))
                    ax1.bar3d(data['row'][i], data['col'][i], pos, 1, 1, data['ave_re'][i], color=np.array([1, 0.9, 0.9*color]), **kwargs)
                #ax.bar3d(data['row'], data['col'], 0, 1, 1, data['ave_re'], shade=True, cmap='rainbow')
                ax1.set_xlabel('Row')
                ax1.set_ylabel('Col')
                ax1.set_zlabel('ave_re')

                ax1.set_xticks([i for i in range(row)])
                ax1.set_yticks([i for i in range(col)])

                if min(data['ave_re']) < 0:
                    zmin = 1.2 * min(data['ave_re'])
                else:
                    zmin = 0.8 * min(data['ave_re'])
                
                if max(data['ave_re']) < 0:
                    zmax = 0.8 * max(data['ave_re'])
                else:
                    zmax = 1.2 * max(data['ave_re'])

                ax1.set_zlim(zmin, zmax)

                # ttest
                for i in range(len(data['row'].index)):
                    color = (data['ttest'][i] - min(data['ttest']))/(max(data['ttest'])-min(data['ttest']))
                    ax2.bar3d(data['row'][i], data['col'][i], pos, 1, 1, data['ttest'][i], color=np.array([1, 0.9, 0.9*color]), **kwargs)
                #ax.bar3d(data['row'], data['col'], 0, 1, 1, data['ave_re'], shade=True, cmap='rainbow')
                ax2.set_xlabel('Row')
                ax2.set_ylabel('Col')
                ax2.set_zlabel('ttest')

                ax2.set_xticks([i for i in range(row)])
                ax2.set_yticks([i for i in range(col)])

                if min(data['ttest']) < 0:
                    zmin = 1.2 * min(data['ttest'])
                else:
                    zmin = 0.8 * min(data['ttest'])
                
                if max(data['ttest']) < 0:
                    zmax = 0.8 * max(data['ttest'])
                else:
                    zmax = 1.2 * max(data['ttest'])

                ax2.set_zlim(zmin, zmax)
                
                # alpha
                for i in range(len(data['row'].index)):
                    color = (data['alpha'][i] - min(data['alpha']))/(max(data['alpha'])-min(data['alpha']))
                    ax3.bar3d(data['row'][i], data['col'][i], pos, 1, 1, data['alpha'][i], color=np.array([1, 0.9, 0.9*color]), **kwargs)
                #ax.bar3d(data['row'], data['col'], 0, 1, 1, data['ave_re'], shade=True, cmap='rainbow')
                ax3.set_xlabel('Row')
                ax3.set_ylabel('Col')
                ax3.set_zlabel('alpha')
                
                ax3.set_xticks([i for i in range(row)])
                ax3.set_yticks([i for i in range(col)])

                if min(data['alpha']) < 0:
                    zmin = 1.2 * min(data['alpha'])
                else:
                    zmin = 0.8 * min(data['alpha'])
                
                if max(data['alpha']) < 0:
                    zmax = 0.8 * max(data['alpha'])
                else:
                    zmax = 1.2 * max(data['alpha'])

                ax3.set_zlim(zmin, zmax)

                # alpha tvalue
                for i in range(len(data['row'].index)):
                    color = (data['alpha_tvalue'][i] - min(data['alpha_tvalue']))/(max(data['alpha_tvalue'])-min(data['alpha_tvalue']))
                    ax4.bar3d(data['row'][i], data['col'][i], pos, 1, 1, data['alpha_tvalue'][i], color=np.array([1, 0.9, 0.9*color]), **kwargs)
                #ax.bar3d(data['row'], data['col'], 0, 1, 1, data['ave_re'], shade=True, cmap='rainbow')
                ax4.set_xlabel('Row')
                ax4.set_ylabel('Col')
                ax4.set_zlabel('alpha_tvalue')
                
                ax4.set_xticks([i for i in range(row)])
                ax4.set_yticks([i for i in range(col)])

                if min(data['alpha_tvalue']) < 0:
                    zmin = 1.2 * min(data['alpha_tvalue'])
                else:
                    zmin = 0.8 * min(data['alpha_tvalue'])
                
                if max(data['alpha_tvalue']) < 0:
                    zmax = 0.8 * max(data['alpha_tvalue'])
                else:
                    zmax = 1.2 * max(data['alpha_tvalue'])

                ax4.set_zlim(zmin, zmax)

                plt.show()
        
        elif pattern == 'scatter':
            if alpha == False:
                fig1, ax1 = plt.subplots(1, 1, figsize=figsize, dpi=500)
                sns.scatterplot(data=data, x='row', y='ave_re', style='col', hue='col', ax=ax1, legend='full', **kwargs)
                ax1.set_xticks([i for i in range(row-1)])
                
                fig2, ax2 = plt.subplots(1, 1, figsize=figsize, dpi=500)
                sns.scatterplot(data=data, x='row', y='ttest', style='col', hue='col', ax=ax2, legend='full', **kwargs)
                ax2.set_xticks([i for i in range(row-1)])
                
                fig3, ax3 = plt.subplots(1, 1, figsize=figsize, dpi=500)
                sns.scatterplot(data=data, x='col', y='ave_re', style='row', hue='row', ax=ax3, legend='full', **kwargs)
                ax3.set_xticks([i for i in range(col-1)])

                fig4, ax4 = plt.subplots(1, 1, figsize=figsize, dpi=500)
                sns.scatterplot(data=data, x='col', y='ttest', style='row', hue='row', ax=ax4, legend='full', **kwargs)
                ax4.set_xticks([i for i in range(col-1)])

            elif alpha == True:
                fig1, ax1 = plt.subplots(1, 1, figsize=figsize, dpi=500)
                sns.scatterplot(data=data, x='row', y='ave_re', style='col', hue='col', ax=ax1, legend='full', **kwargs)
                ax1.set_xticks([i for i in range(row-1)])

                fig2, ax2 = plt.subplots(1, 1, figsize=figsize, dpi=500)
                sns.scatterplot(data=data, x='row', y='ttest', style='col', hue='col', ax=ax2, legend='full', **kwargs)
                ax2.set_xticks([i for i in range(row-1)])
                
                fig3, ax3 = plt.subplots(1, 1, figsize=figsize, dpi=500)
                sns.scatterplot(data=data, x='col', y='ave_re', style='row', hue='row', ax=ax3, legend='full', **kwargs)
                ax3.set_xticks([i for i in range(col-1)])
                
                fig4, ax4 = plt.subplots(1, 1, figsize=figsize, dpi=500)
                sns.scatterplot(data=data, x='col', y='ttest', style='row', hue='row', ax=ax4, legend='full', **kwargs)
                ax4.set_xticks([i for i in range(col-1)])

                fig5, ax5 = plt.subplots(1, 1, figsize=figsize, dpi=500)
                sns.scatterplot(data=data, x='row', y='alpha', style='col', hue='col', ax=ax5, legend='full', **kwargs)
                ax5.set_xticks([i for i in range(row-1)])
                
                fig6, ax6 = plt.subplots(1, 1, figsize=figsize, dpi=500)
                sns.scatterplot(data=data, x='row', y='alpha_tvalue', style='col', hue='col', ax=ax6, legend='full', **kwargs)
                ax6.set_xticks([i for i in range(row-1)])
                
                fig7, ax7 = plt.subplots(1, 1, figsize=figsize, dpi=500)
                sns.scatterplot(data=data, x='col', y='alpha', style='row', hue='row', ax=ax7, legend='full', **kwargs)
                ax7.set_xticks([i for i in range(col-1)])

                fig8, ax8 = plt.subplots(1, 1, figsize=figsize, dpi=500)
                sns.scatterplot(data=data, x='col', y='alpha_tvalue', style='row', hue='row', ax=ax8, legend='full', **kwargs)
                ax8.set_xticks([i for i in range(col-1)])











    








'''
Time Series Analysis 时间序列分析

'''

from pandas import DataFrame
from numpy import ndarray
import numpy as np
from statsmodels.api import OLS
from statsmodels.api import add_constant
import prettytable as pt
from numpy.linalg import inv
from scipy.stats import f


from .adjust import newey_west_t



class TS_regress() :
    '''
    This class is designed for time series regression, 
        r_{i,t} = \beta_if_t + \epsilon_{i,t}
    to obtain the beta for each asset.
    '''
    def __init__(self, list_y:DataFrame, factor:DataFrame) :
        '''
        This function initializes the object.
        input :
            list_y (list/DataFrame): The return matrix with i rows and t columns.
            factor (ndarray or DataFrame): The factor risk premium return series.
        '''

        if type(list_y).__name__ == 'DataFrame':
            self._name_y = list(list_y.columns)
            self.list_y = [np.array(list_y.iloc[:, i]) for i in range(len(list_y.columns))]
        elif type(list_y).__name__ == 'list':
            self._name_y = None
            self.list_y = list_y
        else:
            raise IOError
        
        if type(factor).__name__ == 'DataFrame':
            self._name_factor = list(factor.columns)
            self.factor = np.array(factor)
        elif type(factor).__name__ == 'ndarray':
            self._name_factor = None
            self.factor = factor
        else:
            raise IOError

    def ts_regress(self, newey_west: bool=True, **kwargs) :
        '''
        This function is for conducting the time series regression.
        input :
            newey_west  (boolean): conduct the newey_west adjustment or not.

        output :
            self.alpha (list): The regression alpha.
            self.e_mat (ndarray): The error matrix.
        
        Example:
        from statsmodels.base.model import Model
        from statsmodels.tools.tools import add_constant
        from EAP.time_series_regress import TS_regress

        X = np.random.normal(loc=0.0, scale=1.0, size=(2000,10))
        y_list = []
        for i in range(10) :
            b = np.random.uniform(low=0.1, high=1.1, size=(10,1))
            e = np.random.normal(loc=0.0, scale=1.0, size=(2000,1))
            y = X.dot(b) + e 
            y_list.append(y)

        re = TS_regress(y_list, X)
        '''

        length = len(self.list_y)
        try :            
            r, c = np.shape(self.factor)
        except:
            r = len(self.factor)
            c = 0
        self.alpha = []
        self.e_mat = np.zeros((r, length))
        self.t_value = np.zeros((length, c+1))
        self.p_value = np.zeros((length, c+1))
        self.params_table = np.zeros((length, c+1))
        self.r_square = np.zeros((length, 1))
        self.adj_r = np.zeros((length, 1))

        for i in range(length) :
            result = OLS(self.list_y[i], add_constant(self.factor)).fit()
            params = result.params
            self.params_table[i, :] = params
            residue = result.resid
            self.alpha.append(params[0])
            self.e_mat[:, i] = residue
            self.r_square[i, :] = result.rsquared
            self.adj_r[i, :] = result.rsquared_adj
            if newey_west == True :
                self.t_value[i, :], self.p_value[i, :] = newey_west_t(self.list_y[i], add_constant(self.factor), params=result.params, constant=False,**kwargs)
            elif newey_west == False:
                self.t_value[i, :], self.p_value[i, :] = result.tvalues, result.pvalues
        
        self.alpha = np.array(self.alpha)

        return self.alpha, self.e_mat

    def fit(self, **kwargs) :
        '''
        Fit model 拟合模型
        '''

        self.ts_regress(**kwargs)
    
    def summary(self) : 
        '''
        Summary 总结
        This function summarize the result, including the GRS test.
        '''

        r, c = np.shape(self.factor)
        length = len(self.list_y)
        table = pt.PrettyTable()
        if self._name_factor is None:
            table.field_names = ['Variable', 'alpha'] + ['factor '+str(i) for i in range(c)]
        else:
            table.field_names = ['Variable', 'alpha'] + self._name_factor
        for i in range(length):
            if self._name_y is None:
                table.add_row([str(i)] + list(np.around(self.params_table[i, :], decimals=4)))
            else:
                table.add_row([self._name_y[i]] + list(np.around(self.params_table[i, :], decimals=4)))
            table.add_row(['t-value'] + list(np.around(self.t_value[i, :], decimals=3)))
            table.add_row(['p-value'] + list(np.around(self.p_value[i, :], decimals=3)))
            table.add_row(['R2'] + list(np.around(self.r_square[i, :], decimals=3)) + ['' for num in range(c)])
            table.add_row(['Adj_R2'] + list(np.around(self.adj_r[i, :], decimals=3)) + ['' for num in range(c)])
              
        print(table)

        grs_stats, grs_p = self.grs()
        print("----------------------------------- GRS Test --------------------------------\n")
        print("GRS Statistics:", np.around(grs_stats[0, 0], decimals=3), "GRS p_value:", np.around(grs_p[0, 0], decimals=3))
        print("-----------------------------------------------------------------------------\n")

    def grs(self) :
        '''
        GRS test GRS检验
        This function conducts the GRS test.
        output :
            grs_stats (list):* The GRS statistics.
            p_value (list):* The p_value.
        '''

        length = len(self.list_y)
        r, c = np.shape(self.factor)
        factor_mean = np.expand_dims(np.mean(self.factor, axis=0), axis=0)
        
        temp_factor = self.factor - factor_mean
        sigma_factor_mat = 1 / r * temp_factor.T.dot(temp_factor)
        error_mat = 1 / r * self.e_mat.T.dot(self.e_mat)

        grs_stats = (r - length - c) / (length * (1 + factor_mean.dot(inv(sigma_factor_mat)).dot(factor_mean.T)))\
                    * self.alpha.dot(inv(error_mat)).dot(self.alpha.T)
        p_value = 1 - f.cdf(grs_stats, length, r-length-c)

        return grs_stats, p_value



'''
Test time frequency
'''
# %%
import sys,os

sys.path.append(os.path.abspath(".."))

# %%
from time_frequency import wavelet 
import numpy as np

s = np.random.normal(loc=0.0, scale=1.0, size=(5001))

# %%
wl = wavelet(s)
wave_dec = wl._decompose('db2', mode='multi', level=6)

# %%
pick_series = wl._pick_details()

print(np.shape(pick_series))

# %%
wave_rec = wl._rebuild()


# %%
import matplotlib.pyplot as plt

for i in range(len(wave_rec)):
    plt.subplot(len(wave_rec), 1, i+1)
    plt.plot(wave_rec[i])

# %%
w2 = wavelet(s)
a = w2.fit('db2', mode='multi', level=6, output=True)
plt.plot(np.sum(a, axis=0))
plt.plot(s)

# %%
import numpy as np

factors = np.random.normal(loc=0.0, scale=1.0, size=(4000, 5))
e = np.random.normal(loc=0.0, scale=1.0, size=(4000, 1))
beta = np.random.normal(loc=0.0, scale=1.0, size=(5, 1))

r = factors.dot(beta) + e

# %
from time_frequency import wavelet_pricing as wlp

model = wlp(r, factors)
result = model.fit(wave='db2', mode='multi', level=6)
model.summary()

# %%
import pywt

s = np.random.normal(loc=0.0, scale=1.0, size=(5000))
plt.plot(s)

# %
coef = pywt.wavedec(s, wavelet='db2', level=6)

# % reconstruct
def pick_details(coef):
    '''
    input :
        wave_dec : 
    '''
    import numpy as np

    pick_series = list()    
    for i in range(len(coef)):
        temp_dec = list()
        for j in range(len(coef)):
            if i == j :
                temp_dec.append(coef[j])
            elif i != j :
                temp_dec.append(np.zeros_like(coef[j]))

        pick_series.append(temp_dec)
        
    return pick_series

s_pick = pick_details(coef)
s_rec = list()
for i in range(len(s_pick)):
    s_rec.append(pywt.waverec(s_pick[i], 'db2'))

# %%

s_rec_sum = np.sum(s_rec, axis=0)
plt.plot(s_rec_sum)

# %
s_rec_resid = s - s_rec_sum
plt.plot(s_rec_resid)
# %%



'''
Test time_series_regress
'''
# %%
import sys,os
from numpy import testing

from statsmodels.base.model import Model
from statsmodels.tools.tools import add_constant
sys.path.append(os.path.abspath(".."))

from time_series_regress import TS_regress

# %%
def test_TS_regress() :
    import numpy as np

    X = np.random.normal(loc=0.0, scale=1.0, size=(2000,10))
    y_list = []
    for i in range(10) :
        b = np.random.uniform(low=0.1, high=1.1, size=(10,1))
        e = np.random.normal(loc=0.0, scale=1.0, size=(2000,1))
        y = X.dot(b) + e 
        y_list.append(y)

    re = TS_regress(y_list, X)
    re.fit()
    re.summary()
    print(re.grs())

test_TS_regress()

# %%



'''
Factor model for industry 商用因子模型
'''

import numpy as np
import matplotlib.pyplot as plt
from prettytable import PrettyTable

from .portfolio_analysis import Univariate

class Portfolio(Univariate):
    '''
    construct portfolio and backtest 
    '''
    def __init__(self, sample):
        super().__init__(sample)


    def divide_token(self):
        stocklist = dict()
        group_time = super().divide_by_time(self.sample)
        self._average_group_by_time = super().average_by_time()
        self._average = self._average_group_by_time.mean(axis=1)

        for t_time in range(len(group_time)):
            group = group_time[t_time]
            if self._weight == False:
                token = group[:, 3]
            elif self._weight == True:
                token = group[:, 4]
            else:
                return IOError
    
            label = self._label[t_time]

            stocklist_eachtime = dict()

            for n_group in range(self.number+1):
                location = np.where(label==n_group)
                stocklist_eachtime[n_group] = token[location]            
            
            stocklist[self._time[t_time]] = stocklist_eachtime


        self.stocklist = stocklist
        return stocklist
    
    def extract_portfolio(self, group_num):
        '''
        extract portfolio return series
        '''
        return self._average_group_by_time[group_num, :]
    
    def plot(self, group_num):
        '''
        plot portfolio return
        '''
        ax = plt.plot(self._time, self._average_group_by_time[group_num, :], label='Group'+str(group_num+1))
        return ax
    
    def plot_all(self):
        '''
        plot all portfolio return
        '''
        for i in range(self.number+1):
            self.plot(i)
        
        plt.title('Portfolio Return')
        plt.legend()

    def group_turnover(self, group_num):
        '''
        calculate group turnover
        '''
        turnover_each_time = np.zeros(len(self._time) - 1)
        stocklist_previous = self.stocklist[self._time[0]][group_num]

        for t_time in range(len(self._time)-1):
            stocklist_eachtime = self.stocklist[self._time[t_time + 1]][group_num]
            stocklist_common = set(stocklist_eachtime) & set(stocklist_previous)
            turnover_each_time[t_time] = 1 - len(stocklist_common) / len(stocklist_eachtime)

            # upgrade stock_previous
            stocklist_previous = stocklist_eachtime
        
        return turnover_each_time

    def turnover(self):
        '''
        calculate all turnover : 计算全组换手率
        '''
        turnover_allgroup = np.zeros((self.number+1, len(self._time)-1))
        for num in range(self.number + 1):
            turnover_allgroup[num, :] = self.group_turnover(num)
        
        self._turnover_allgroup = turnover_allgroup
        
        return turnover_allgroup

    def compound_return(self):
        '''
        compound return : 累计收益率
        '''
        average_group_by_time_add_one = np.c_[np.zeros((self.number+1, 1)), self._average_group_by_time]
        self._group_compound_return = np.cumprod(1 + average_group_by_time_add_one, axis=1)

        return self._group_compound_return
    
    def plot_compound_return(self):
        '''
        plot compound return
        '''
        for group_num in range(self.number+1):
            plt.plot(np.r_[self._time[0], self._time], self._group_compound_return[group_num, :], label='Group'+str(group_num+1))

        plt.title('Compound Return')        
        plt.legend()

    def drawdown(self, group_num):
        '''
        calculate drawdown : 计算回撤
        '''
        group_compound_return = self.compound_return()[group_num, :]
        group_max = group_compound_return[0]
        temp_drawdown = 0.0
        group_drawdown = np.zeros((len(group_compound_return), 1))

        for t_time in range(len(group_compound_return)):
            if group_compound_return[t_time] < group_max:
                group_drawdown[t_time] = (group_compound_return[t_time] - group_max) / group_max
            else:
                group_drawdown[t_time] = 0.0
                group_max = group_compound_return[t_time]

            temp_drawdown = group_drawdown[t_time]

        return group_drawdown

    def maxdrawdown(self):
        '''
        calculate MAX drawdown : 计算最大回撤
        '''
        all_drawndown = np.zeros(np.shape(self._group_compound_return))

        for num in range(self.number+1):
            all_drawndown[num, :] = self.drawdown(num)[:, 0]
        
        self._all_drawdown = all_drawndown
        self._max_drawdown = self._all_drawdown.min(axis=1)

        return self._max_drawdown        
    
    def plot_drawdown(self):
        '''
        plot drawdown : 画出回撤
        '''
        for group_num in range(self.number+1):
            plt.plot(np.r_[self._time[0], self._time], self._all_drawdown[group_num, :], label='Group'+str(group_num+1))
        
        plt.plot('Max DrawDown')
        plt.legend()

    def sharpe_ratio(self):
        '''
        calculate sharp ratio
        '''
        self._group_sharpe_ratio = self._average_group_by_time.mean(axis=1) / self._average_group_by_time.std(axis=1)
        
        return self._group_sharpe_ratio
    
    def backtest(self):
        '''
        Backtest : 回测
        '''
        
        self.divide_token()
        self.compound_return()
        self.turnover()
        self.maxdrawdown()
        self.sharpe_ratio()

    def backtest_summary(self, decimals=3):
        '''
        Backtest summary : 回测结果
        '''
        backtest_table = PrettyTable(["Statistics", "Value"])
        
        for group in range(self.number+1):
            backtest_table.add_row(["Group  "+str(group+1), " "])
            backtest_table.add_row(["Average return", np.around(self._average[group], decimals=decimals)])
            backtest_table.add_row(["compounded return", np.around(self._group_compound_return[group, -1], decimals=decimals)])
            backtest_table.add_row(["Turnover", np.around(self._turnover_allgroup.mean(axis=1)[group], decimals=decimals)])
            backtest_table.add_row(["MaxDrawdown", np.around(self._max_drawdown[group], decimals=decimals)])
            backtest_table.add_row(["Sharpte Ratio", np.around(self._group_sharpe_ratio[group], decimals=decimals)])

            if group != self.number:
                backtest_table.add_row(["   ", "   "])

        print(backtest_table)
